# 3rd Party
from flask import current_app, make_response

# Fastlane
from fastlane.helpers import dumps


def return_error(msg, operation, status=500, logger=None):
    if logger is None:
        logger = current_app.bind(operation=operation)

    logger.error(msg)

    return make_response(dumps({"error": msg, "operation": operation}), status)



# Standard Library
import re
from datetime import timedelta

REGEX = re.compile(r"((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?")


def parse_time(time_str):
    if time_str is None:
        return None

    parts = REGEX.match(time_str)

    if not parts:
        return None

    parts = parts.groupdict()
    time_params = {}

    for (name, param) in parts.items():
        if param:
            time_params[name] = int(param)

    return timedelta(**time_params)



"""   isort:skip_file """

# 3rd Party
from gevent import monkey

# Must be before other imports
monkey.patch_all()  # isort:skip

# Fastlane
from fastlane.api.app import Application  # NOQA pylint: disable=wrong-import-position
from fastlane.config import Config  # NOQA pylint: disable=wrong-import-position


class APIHandler:
    def __init__(self, click, host, port, config, log_level):
        self.config_path = config
        self.config = None
        self.click = click
        self.host = host
        self.port = port
        self.log_level = log_level

        self.load_config()

    def load_config(self):
        # self.click.echo(f'Loading configuration from {self.config_path}...')
        self.config = Config.load(self.config_path)

    def __call__(self):
        app = Application(self.config, self.log_level)
        app.logger.info(
            "fastlane is runnning.",
            host=self.host,
            port=self.port,
            environment=self.config.ENV,
        )
        app.run(self.host, self.port)



# Standard Library
import time
from uuid import uuid4

# 3rd Party
import rq
from rq import Connection, Worker

# Fastlane
from fastlane.api.app import Application
from fastlane.config import Config
from fastlane.worker.job import enqueue_missing_monitor_jobs
from fastlane.worker.scheduler import QueueScheduler


class WorkerHandler:
    def __init__(
        self, click, worker_id, jobs, monitor, notify, webhooks, config, log_level
    ):
        self.config_path = config
        self.config = None
        self.click = click
        self.worker_id = worker_id
        self.log_level = log_level
        self.queues = []

        if jobs:
            self.queues.append("jobs")

        if monitor:
            self.queues.append("monitor")

        if notify:
            self.queues.append("notify")

        if webhooks:
            self.queues.append("webhooks")

        self.load_config()

    def load_config(self):
        # self.click.echo(f'Loading configuration from {self.config_path}...')
        self.config = Config.load(self.config_path)

    def __call__(self):
        # self.click.echo(
        # f'Running fastlane worker processing queues {",".join(self.queues)}.')
        app = Application(self.config, self.log_level, auto_connect_db=True)
        app.logger.info(
            f'Running fastlane worker processing queues {",".join(self.queues)}.'
        )
        interval = app.config["WORKER_SLEEP_TIME_MS"] / 1000.0

        with app.app.app_context():
            worker_kw = dict(connection=app.app.redis)

            if self.worker_id is None:
                app.logger.warn(
                    "The worker id was not set for this worker and a random one will be used."
                )
                self.worker_id = str(uuid4())

            app.logger = app.logger.bind(worker_id=self.worker_id, queues=self.queues)
            app.app.logger = app.logger
            worker_kw["name"] = self.worker_id

            worker = Worker(self.queues, **worker_kw)

            schedulers = {}

            with Connection(app.app.redis):
                for queue in self.queues:
                    schedulers[queue] = QueueScheduler(queue, app=app.app)
                app.schedulers = schedulers

                # app.logger.debug('Processing enqueued items...')

                try:
                    while True:
                        for queue in self.queues:
                            # app.logger.debug("Processing scheduler...", queue=queue)
                            schedulers[queue].move_jobs()

                        # app.logger.debug('Processing queues...')
                        worker.work(burst=True)
                        enqueue_missing_monitor_jobs(app.app)
                        time.sleep(interval)
                except rq.worker.StopRequested:
                    app.logger.info("Worker exiting gracefully.")

                    return



# Standard Library
from datetime import datetime

# 3rd Party
import croniter
import pkg_resources
from flask import Blueprint, current_app, jsonify, url_for

# Fastlane
from fastlane.models import Job, Task

bp = Blueprint("status", __name__, url_prefix="/status")  # pylint: disable=invalid-name


@bp.route("/", methods=("GET",))
def status():
    executor = current_app.executor
    version = pkg_resources.get_distribution("fastlane").version
    metadata = {"hosts": [], "containers": {"running": []}}

    containers = executor.get_running_containers()

    for host, port, container_id in containers["running"]:
        metadata["containers"]["running"].append(
            {"host": host, "port": port, "id": container_id}
        )

    metadata["hosts"] = [] + containers["available"] + containers["unavailable"]
    metadata["queues"] = {"jobs": {}, "monitor": {}, "error": {}}

    for queue in ["jobs", "monitor", "error"]:
        jobs_queue_size = current_app.redis.llen(f"rq:queue:{queue}")
        metadata["queues"][queue]["length"] = jobs_queue_size

    metadata["tasks"] = {"count": Task.objects.count()}

    metadata["jobs"] = {"count": Job.objects.count()}

    metadata["jobs"]["scheduled"] = []
    scheduled_jobs = Job.objects(scheduled=True).all()

    metadata["fastlane"] = {
        "version": version,
        "executor": current_app.config["EXECUTOR"],
    }

    for job in scheduled_jobs:
        j = job.to_dict(include_executions=False)

        itr = croniter.croniter(job.metadata["cron"], datetime.utcnow())
        j["nextScheduledAt"] = itr.get_next(datetime).isoformat()

        task_id = job.task.task_id

        job_url = url_for(
            "task.get_job", task_id=task_id, job_id=str(job.id), _external=True
        )
        j["url"] = job_url

        stop_job_url = url_for(
            "task.stop_job", task_id=task_id, job_id=str(job.id), _external=True
        )
        j["stopUrl"] = stop_job_url

        task_url = url_for("task.get_task", task_id=task_id, _external=True)
        del j["taskId"]
        j["task"] = {"id": task_id, "url": task_url}

        metadata["jobs"]["scheduled"].append(j)

    return jsonify(metadata), 200



# Standard Library
import logging
import sys
from json import loads

# 3rd Party
import rq_dashboard
import structlog
from flask import Flask
from flask_cors import CORS
from flask_redis import FlaskRedis
from flask_redis_sentinel import SentinelExtension
from flask_sockets import Sockets
from gevent import pywsgi
from geventwebsocket.handler import WebSocketHandler
from structlog.processors import (
    JSONRenderer,
    StackInfoRenderer,
    TimeStamper,
    format_exc_info,
)
from structlog.stdlib import add_log_level, add_logger_name, filter_by_level

# Fastlane
import fastlane.api.gzipped as gzipped
import fastlane.api.metrics as metrics
import fastlane.api.rqb as rqb
from fastlane.api.enqueue import bp as enqueue
from fastlane.api.execution import bp as execution_api
from fastlane.api.healthcheck import bp as healthcheck
from fastlane.api.status import bp as status
from fastlane.api.stream import bp as stream
from fastlane.api.task import bp as task_api
from fastlane.models import db


class Application:
    def __init__(self, config, log_level, testing=False, auto_connect_db=True):
        self.config = config
        self.logger = None
        self.log_level = log_level

        self.create_app(testing, auto_connect_db=auto_connect_db)

    def create_app(self, testing, auto_connect_db=True):
        self.app = Flask("fastlane")

        self.testing = testing
        self.app.testing = testing
        self.app.config.from_object(rq_dashboard.default_settings)
        self.app.error_handlers = []

        for key in self.config.items.keys():
            self.app.config[key] = self.config[key]

        self.app.config["ENV"] = self.config.ENV
        self.app.config["DEBUG"] = self.config.DEBUG
        self.app.original_config = self.config
        self.app.log_level = self.log_level
        self.configure_logging()
        self.connect_redis()
        self.connect_queue()

        if auto_connect_db:
            self.connect_db()
        self.load_executor()
        self.load_error_handlers()

        enable_cors = self.app.config["ENABLE_CORS"]

        if (
            isinstance(enable_cors, (str, bytes)) and enable_cors.lower() == "true"
        ) or (isinstance(enable_cors, (bool)) and enable_cors):
            self.app.logger.info(
                f'Configured CORS to allow access from \'{self.app.config["CORS_ORIGINS"]}\'.'
            )
            CORS(self.app)

        metrics.init_app(self.app)
        self.app.register_blueprint(metrics.bp)
        self.app.register_blueprint(healthcheck)
        self.app.register_blueprint(enqueue)
        self.app.register_blueprint(task_api)
        self.app.register_blueprint(execution_api)
        self.app.register_blueprint(status)

        self.app.register_blueprint(gzipped.bp)
        gzipped.init_app(self.app)

        sockets = Sockets(self.app)
        sockets.register_blueprint(stream)

    def configure_logging(self):
        if self.app.testing:
            structlog.reset_defaults()

        disabled = [
            "docker.utils.config",
            "docker.auth",
            "docker.api.build",
            "docker.api.swarm",
            "docker.api.image",
            "rq.worker",
            "werkzeug",
            "requests",
            "urllib3",
        ]

        for logger in disabled:
            log = logging.getLogger(logger)
            log.setLevel(logging.ERROR)
            log.disabled = True
        self.app.logger.disabled = True

        logging.basicConfig(
            level=self.log_level, stream=sys.stdout, format="%(message)s"
        )

        chain = [
            filter_by_level,
            add_log_level,
            add_logger_name,
            TimeStamper(fmt="iso"),
            StackInfoRenderer(),
            format_exc_info,
            JSONRenderer(indent=1, sort_keys=True),
        ]

        logger = logging.getLogger(__name__)

        if self.testing:
            chain = []
            logger = structlog.ReturnLogger()

        log = structlog.wrap_logger(
            logger,
            processors=chain,
            context_class=dict,
            wrapper_class=structlog.stdlib.BoundLogger,
            # cache_logger_on_first_use=True,
        )
        self.logger = log
        self.app.logger = self.logger

    def connect_redis(self):
        self.logger.debug("Connecting to redis...")

        if self.app.testing:
            self.logger.info("Configuring Fake Redis...")
            import fakeredis

            self.app.redis = FlaskRedis.from_custom_provider(fakeredis.FakeStrictRedis)
            self.app.redis.connect = self._mock_redis(True)
            self.app.redis.disconnect = self._mock_redis(False)
            self.app.redis.init_app(self.app)
        elif self.app.config["REDIS_URL"].startswith("redis+sentinel"):
            self.logger.info(
                "Configuring Redis Sentinel...", redis_url=self.app.config["REDIS_URL"]
            )
            redis_sentinel = SentinelExtension()
            redis_connection = redis_sentinel.default_connection
            redis_sentinel.init_app(self.app)
            self.app.redis = redis_connection
        else:
            self.logger.info(
                "Configuring Redis...", redis_url=self.app.config["REDIS_URL"]
            )
            self.app.redis = FlaskRedis()
            self.app.redis.init_app(self.app)

        self.logger.info("Connection to redis successful")

    def connect_queue(self):
        self.app.queue = None
        self.app.register_blueprint(rqb.bp)
        rqb.init_app(self.app)

    def connect_db(self):
        settings = self.app.config["MONGODB_CONFIG"]

        if isinstance(settings, (dict,)):
            self.app.config["MONGODB_SETTINGS"] = settings
        else:
            self.app.config["MONGODB_SETTINGS"] = loads(
                self.app.config["MONGODB_CONFIG"]
            )

        self.logger.info(
            "Connecting to MongoDB...", mongo=self.app.config["MONGODB_SETTINGS"]
        )
        db.init_app(self.app)
        self.logger.info(
            "Connected to MongoDB successfully.",
            mongo=self.app.config["MONGODB_SETTINGS"],
        )

    def load_executor(self):
        name = self.config.EXECUTOR
        parts = name.split(".")
        executor_module = __import__(".".join(parts), None, None, [parts[-1]], 0)

        self.app.executor_module = executor_module

        blueprint = getattr(executor_module, "bp", None)

        if blueprint is not None:
            self.app.register_blueprint(blueprint)

        self.app.executor = self.app.executor_module.Executor(self.app)

    def load_error_handlers(self):
        self.app.error_handlers = []

        for handler_name in self.app.config["ERROR_HANDLERS"]:
            parts = handler_name.split(".")
            obj = __import__(".".join(parts[:-1]), None, None, [parts[-1]], 0)
            obj = getattr(obj, parts[-1])

            self.app.error_handlers.append(obj(self.app))

        self.app.report_error = self.report_error

    def report_error(self, err, metadata=None):
        for handler in self.app.error_handlers:
            handler.report(err, metadata)

    def run(self, host, port):
        server = pywsgi.WSGIServer(
            (host, port), self.app, handler_class=WebSocketHandler
        )
        server.serve_forever()

    def _mock_redis(self, connected):
        def handle():
            self.app.redis._redis_client.connected = connected

        return handle



# 3rd Party
from rq_scheduler import Scheduler


class QueueScheduler:
    def __init__(self, queue_name, app):
        self.app = app
        self.logger = self.app.logger.bind(queue_name=queue_name)
        self.scheduler = Scheduler(queue_name=queue_name, connection=app.redis)

    def move_jobs(self):
        if self.scheduler.acquire_lock():
            try:
                jobs = self.scheduler.get_jobs()
                self.logger.debug(
                    "Lock acquired. Enqueuing scheduled jobs...", jobs=jobs
                )
                self.scheduler.enqueue_jobs()
            finally:
                self.scheduler.remove_lock()
        else:
            self.logger.debug(
                "Lock could not be acquired. Enqueuing scheduled "
                "jobs skipped. Trying again next cycle."
            )



# 3rd Party
from flask import Blueprint
from rq import Queue
from rq_scheduler import Scheduler

bp = Blueprint("rq", __name__)  # pylint: disable=invalid-name


class JobQueue:
    def __init__(self, queue_name, app):
        self.app = app
        self.queue_name = queue_name
        self._queue = None
        self._scheduler = None

    def enqueue_in(self, *args, **kw):
        self.app.logger.info("Scheduling execution for the future.", **kw)

        return self._scheduler.enqueue_in(*args, **kw)

    def enqueue_at(self, *args, **kw):
        self.app.logger.info("Scheduling execution for a specific timestamp.", **kw)

        return self._scheduler.enqueue_at(*args, **kw)

    def enqueue(self, *args, **kw):
        return self.queue.enqueue(*args, **kw)

    @property
    def queue(self):
        if self._queue is None:
            self._queue = Queue(self.queue_name, connection=self.app.redis)

        return self._queue

    @property
    def scheduler(self):
        if self._scheduler is None:
            self._scheduler = Scheduler(queue=self._queue)

        return self._scheduler


def init_app(app):
    for queue_name in ["jobs", "monitor", "notify", "webhooks"]:
        key = queue_name.rstrip("s")
        setattr(app, "%s_queue" % key, JobQueue(queue_name, app))



class HostUnavailableError(RuntimeError):
    def __init__(self, host, port, error):
        msg = f"Connection to host {host}:{port} failed with error: {error}"
        super(HostUnavailableError, self).__init__(msg)
        self.host = host
        self.port = port
        self.error = error
        self.message = f"Connection to host {self.host}:{self.port} failed with error: {self.error}"


class NoAvailableHostsError(RuntimeError):
    pass


class ContainerUnavailableError(RuntimeError):
    pass



# Standard Library
import datetime
from uuid import uuid4

# 3rd Party
import mongoengine.errors
from flask import url_for
from mongoengine import DateTimeField, ListField, ReferenceField, StringField

# Fastlane
from fastlane.models import db


class Task(db.Document):
    created_at = DateTimeField(required=True)
    last_modified_at = DateTimeField(required=True, default=datetime.datetime.now)
    task_id = StringField(required=True)
    jobs = ListField(ReferenceField("Job"))

    meta = {
        "indexes": ["task_id", {"fields": ["$task_id"], "default_language": "english"}]
    }

    def _validate(self):
        errors = {}

        if self.task_id == "":
            errors["task_id"] = mongoengine.errors.ValidationError(
                "Field is required", field_name="task_id"
            )

        if errors:
            message = "ValidationError (%s:%s) " % (self._class_name, self.pk)
            raise mongoengine.errors.ValidationError(message, errors=errors)

    def save(self, *args, **kwargs):
        self._validate()

        if not self.created_at:
            self.created_at = datetime.datetime.now()
        self.last_modified_at = datetime.datetime.now()

        return super(Task, self).save(*args, **kwargs)

    def get_url(self):
        return url_for("task.get_task", task_id=self.task_id, _external=True)

    def to_dict(self):
        res = {
            "taskId": self.task_id,
            "createdAt": self.created_at.isoformat(),
            "lastModifiedAt": self.last_modified_at.isoformat(),
            "url": self.get_url(),
            "jobsCount": len(self.jobs),
        }

        return res

    @classmethod
    def create_task(cls, task_id):
        new_task = cls(task_id=task_id)
        new_task.save()

        return new_task

    @classmethod
    def get_tasks(cls, page=1, per_page=20):
        return cls.objects.paginate(page, per_page)  # pylint: disable=no-member

    @classmethod
    def search_tasks(cls, query, page=1, per_page=20):
        return cls.objects.search_text(query).paginate(page, per_page)

    @classmethod
    def get_by_task_id(cls, task_id):
        if task_id is None or task_id == "":
            raise RuntimeError("Task ID is required and can't be None or empty.")

        return cls.objects(task_id=task_id).no_dereference().first()

    def create_job(self):
        from fastlane.models.job import Job

        job_id = uuid4()
        j = Job(task_id=str(self.task_id), job_id=str(job_id))
        j.task = self
        j.save()

        self.jobs.append(j)
        self.save()

        return j

    def create_or_update_job(self, job_id):
        from fastlane.models.job import Job

        jobs = list(filter(lambda job: str(job.job_id) == job_id, self.jobs))

        if not jobs:
            j = Job(task_id=str(self.task_id), job_id=str(job_id))
            j.task = self
            j.save()
            self.jobs.append(j)
            self.save()
        else:
            j = jobs[0]

        return j



"""   isort:skip_file """

# 3rd Party
from flask_mongoengine import MongoEngine

db = MongoEngine()  # isort:skip pylint: disable=invalid-name

from fastlane.models.task import (  # NOQA pylint: disable=unused-import,wrong-import-position
    Task,
)
from fastlane.models.job import (  # NOQA pylint: disable=unused-import,wrong-import-position
    JobExecution,
    Job,
)



# Standard Library
from datetime import datetime
from uuid import uuid4

# 3rd Party
from flask import Blueprint, current_app, g, request

bp = Blueprint("metrics", __name__)  # pylint: disable=invalid-name


def init_app(app):
    def start_timer():
        request_id = request.headers.get("X-Request-ID", str(uuid4()))
        g.logger = current_app.logger.bind(request_id=request_id)
        g.request_id = request_id
        g.start = datetime.now()

    app.before_request(start_timer)

    def log_request(response):
        if request.path == "/favicon.ico":
            return response

        now = datetime.now()
        duration = int(round((now - g.start).microseconds / 1000, 2))

        user_ip = request.headers.get("X-Forwarded-For", request.remote_addr)
        host = request.host.split(":", 1)[0]

        log_params = {
            "method": request.method,
            "path": request.path,
            "status": response.status_code,
            "duration": duration,
            "ip": user_ip,
            "host": host,
        }

        request_id = g.request_id

        if request_id:
            log_params["request_id"] = request_id

        if response.status_code < 400:
            current_app.logger.info("Request succeeded", **log_params)
        elif response.status_code < 500:
            current_app.logger.info("Bad Request", **log_params)
        else:
            current_app.logger.error("Internal Server Error", **log_params)

        return response

    app.after_request(log_request)
